/* wifi-sim: a brain-dead wifi simulator
 *
 * Copyright (c) 2010, Bob Copeland <me@bobcopeland.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <time.h>

#define NRATES 8
#define MAX_RETRY 4
#define MAX_SNR ARRAY_SIZE(per)/NRATES - 1

#define WARM_UP 5
#define START_PKT_LEN 1000

#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define clamp(a,b,c) (min(max(a,b),c))

static int index_to_rate[] = {
	/* 10, 20, 55, 110, */ 60, 90, 120, 180, 240, 360, 480, 540
};

static int rate_to_per_index[] = {
    540, 480, 360, 240, 180, 120, 90, 60
};


#include "per.h"
#include "mac80211_hwsim.h"

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))

typedef unsigned char u8;

/* true if we are simulating wall clock time */
static bool real_time = false;
/* true if sim stays at 30 dB */
static bool fixed_db = false;
/* true if RC is perfect */
static bool perfect_rc = false;

struct timespec trans_time;

static inline int div_round(int a, int b)
{
    return (a + b-1)/b;
}

static inline int pkt_duration(int len, int rate)
{
    /* preamble + signal + t_sym * n_sym, rate in 100 kbps */
    return 16 + 4 + 4 * div_round((16 + 8 * len + 6) * 10, 4 * rate);
}

void timespec_sub(struct timespec *a, struct timespec *b, struct timespec *res)
{
    res->tv_sec = a->tv_sec - b->tv_sec;
    res->tv_nsec = a->tv_nsec - b->tv_nsec;
    if (res->tv_nsec < 0)
    {
        res->tv_sec--;
        res->tv_nsec += 1000000000;
    }
}

double get_per(int rate, int len, double snr)
{
    int i;

    /* get PER given the snr and rate */
    int snr_index = clamp(snr * 20, 0, MAX_SNR);

    for (i=0; i < ARRAY_SIZE(rate_to_per_index); i++)
    {
        if (rate_to_per_index[i] == rate)
            break;
    }
    if (i == ARRAY_SIZE(rate_to_per_index))
    {
        printf("unusable rate: %d\n", rate);
        return 0.5;
    }

    if (len > 100)
        return per[snr_index * NRATES + i];

    return per_lo[snr_index * NRATES + i];
}

int best_rate(int len, int snr)
{
    double max_tp = 0;
    int max_rate = -1;
    int i;
    int snr_index = clamp(snr * 20, 0, MAX_SNR);

    for (i=0; i < ARRAY_SIZE(rate_to_per_index); i++)
    {
        int rate = rate_to_per_index[i];
        double tp = (rate / 10) * (1.0 - per[snr_index * NRATES + i]);
        if (tp > max_tp) {
            max_rate = rate;
            max_tp = tp;
        }
    }
    if (max_rate > 0)
    {
        for (i=0; i < ARRAY_SIZE(index_to_rate); i++)
            if (max_rate == index_to_rate[i])
                return i;
    }
    return -1;
}

bool pass_packet(int rate, int len, int snr)
{
	double per = get_per(rate, len, snr);
	return (drand48() > per);
}

#define PKT_PER_DB 10
void handle_packet(int fd, int fd2, struct timespec *start_time, bool print)
{
	int len;
    int i, j;
	char packet[3000];
	struct mac80211_hwsim_tx_packet *pkt;
	struct mac80211_hwsim_tx_status status;
    double snr = 30;
    double irate = 0, erate = 0;
    int success_rate = 0;
    struct timespec now_time;
    struct timespec elapsed;

    static bool has_started = false;

	len = read(fd, packet, sizeof(packet));

	if (len <= 0) {
		printf("huh, select must be broken %d\n", len);
		usleep(1000);
		return;
	}

    /* wait for at least one large packet before reducing the SNR */
    clock_gettime(CLOCK_MONOTONIC, &now_time);
    if (!has_started)
    {
        if (len > START_PKT_LEN)
        {
            has_started = true;
            printf("--MARK--\n");
        }
        else
            memcpy(start_time, &now_time, sizeof(now_time));
    }
    timespec_sub(&now_time, start_time, &elapsed);

    if (!fixed_db)
    {
        /* every 2 seconds, snr gets dropped by .1 dB */
        if (elapsed.tv_sec > WARM_UP)
            snr -= .1 * ((double) elapsed.tv_sec - WARM_UP) / 2;
    }

	pkt = (struct mac80211_hwsim_tx_packet *) packet;

	memset(&status, 0, sizeof(status));

    /* set up all unused rates to be -1 */
    for (i=0; i < ARRAY_SIZE(status.rates); i++)
        status.rates[i].idx = -1;

    int slot_time = 9;
    int sifs = 16;
    int difs = 2 * slot_time + sifs;
    int cw_min = 15;
    int cw_max = 1023;
    int cw = cw_min;
    int ack_time_usec = pkt_duration(14, index_to_rate[0]) + sifs;
    int sleep_amt = 0;

    if (perfect_rc)
    {
        pkt->rates[0].idx = best_rate(len, snr);
        pkt->rates[1].idx = -1;
    }

    /* try each rate count times until ACKed */
    for (i=0; i < MAX_RETRY && !status.ack; i++)
    {
	    status.rates[i].idx = pkt->rates[i].idx;

        /* no more rates in MRR */
        if (pkt->rates[i].idx < 0)
            break;

        for (j = 0; j < pkt->rates[i].count; j++)
        {
	        int rate = index_to_rate[pkt->rates[i].idx];
	        status.rates[i].count++;

            /* sending takes some time for iperf testing, needs hires timers */
            /* discount elapsed time from previous attempt */
            sleep_amt += difs + pkt_duration(pkt->pkt_len, rate);
            if (j > 0) {
                sleep_amt += (cw * slot_time) / 2;
                cw = (cw << 1) + 1;
                if (cw > cw_max)
                    cw = cw_max;
            }
            /* try "sending" it */
            if (perfect_rc)
                status.ack = true;
            else
	            status.ack = pass_packet(rate, pkt->pkt_len, snr);

            if (status.ack)
            {
                /* skip ack wait for ack packets/mgmt frames... */
                if (pkt->pkt_len > 100)
                    sleep_amt += ack_time_usec;
                break;
            }
        }
    }
    if (real_time)
    {
        struct timespec sleep_time = { .tv_sec = 0 }, rem_time;

        clock_gettime(CLOCK_MONOTONIC, &now_time);
        timespec_sub(&now_time, &trans_time, &elapsed);

        sleep_time.tv_nsec = 1000 * sleep_amt - elapsed.tv_nsec;

        if (sleep_time.tv_nsec > 0)
            nanosleep(&sleep_time, &rem_time);

        clock_gettime(CLOCK_MONOTONIC, &trans_time);
    }

    /* compute instantaneous rate */
    if (status.ack)
    {
        for (i=0; i < ARRAY_SIZE(status.rates) && status.rates[i].idx >= 0; i++)
        {
	        int rate = index_to_rate[status.rates[i].idx];
            irate += status.rates[i].count / (rate / 10.0);
            success_rate = rate / 10;
        }
        irate = 1.0 / irate;
        erate =  (8.0 * pkt->pkt_len) / sleep_amt;
    }

    if (print)
    {
        timespec_sub(&now_time, start_time, &elapsed);
        printf("[%ld.%09ld] ",  elapsed.tv_sec, elapsed.tv_nsec);
        printf ("snr %f ack %d rate %d %d  %d %d  %d %d  %d %d "
                "status %d %d  %d %d  %d %d  %d %d irate %f %f %d\n",
                snr, status.ack,
                pkt->rates[0].idx, pkt->rates[0].count,
                pkt->rates[1].idx, pkt->rates[1].count,
                pkt->rates[2].idx, pkt->rates[2].count,
                pkt->rates[3].idx, pkt->rates[3].count,
                status.rates[0].idx, status.rates[0].count,
                status.rates[1].idx, status.rates[1].count,
                status.rates[2].idx, status.rates[2].count,
                status.rates[3].idx, status.rates[3].count,
                irate, erate, success_rate);
    }

    /* successful send, hand it to other side */
	if (status.ack)
		write(fd2, pkt->pkt_data, pkt->pkt_len);

	status.cookie = pkt->cookie;
	ioctl(fd, MAC80211_HWSIM_IOCTL_STATUS, &status);
}

#define HWSIM_DEV "/dev/mac80211_hwsim_control"
int main(int argc, char *argv[])
{
	int fd;
	int fd2;
    int i;
	fd_set readfds, tmpfds, nofds, dummy;
    struct timespec start_time;

    for (i=1; i < argc; i++)
    {
        if (strcmp(argv[i], "rt") == 0)
            real_time = true;
        else if (strcmp(argv[i], "fx") == 0)
            fixed_db = true;
        else if (strcmp(argv[i], "pr") == 0)
            perfect_rc = true;
    }

    if (real_time)
        printf("Sleeping to simulate real-time TX\n");

	fd = open(HWSIM_DEV, O_RDWR);
	fd2 = open(HWSIM_DEV, O_RDWR);

	if (fd < 0 || fd2 < 0) {
		perror(HWSIM_DEV);
		exit(1);
	}

	FD_ZERO(&readfds);
	FD_ZERO(&nofds);

	FD_SET(fd, &readfds);
	FD_SET(fd2, &readfds);

    clock_gettime(CLOCK_MONOTONIC, &start_time);
    memcpy(&trans_time, &start_time, sizeof(start_time));
	while (1)
	{
		tmpfds = readfds;
		dummy = nofds;

		select(fd2 + 1, &tmpfds, &dummy, &dummy, NULL);
		if (FD_ISSET(fd, &tmpfds))
		{
			handle_packet(fd, fd2, &start_time, true);
		}
		if (FD_ISSET(fd2, &tmpfds))
		{
			handle_packet(fd2, fd, &start_time, false);
		}
	}
}

